import os
from pprint import pformat
from typing import Tuple, Dict

import pandas as pd
import toml

from sevivi.config import (
    SensorConfig,
    ManuallySynchronizedSensorConfig,
    JointSynchronizedSensorConfig,
    ImuSynchronizedSensorConfig,
    CameraImuVideoConfig,
    KinectVideoConfig,
    RawVideoConfig,
    OpenPoseVideoConfig,
    StackingDirection,
    PlottingMethod,
    Config,
    VideoConfig,
    RenderConfig,
    VideoImuCaptureAppVideoConfig,
)
from sevivi.log import logger

logger = logger.getChild("config_reader")


def merge_config_files(config_file_paths: Tuple[str, ...]) -> Dict:
    config_dict = {}

    for path in config_file_paths:
        if not os.path.exists(path):
            raise FileNotFoundError(f"Missing configuration file {path}")
        else:
            update = toml.load(path)

            # In theory we could manually merge the video configurations;
            # However, I can't really think of a use case. The behaviour for
            # sensors is to just add all configured sensors together, which
            # is not wanted for video config.
            if "video" in update and "video" in config_dict:
                logger.info("Only the latest video config is used")
                del config_dict["video"]

            deep_update(config_dict, update)
    return config_dict


def read_configs(config_file_paths: Tuple[str, ...]) -> Config:
    if config_file_paths is None or len(config_file_paths) == 0:
        raise ValueError(
            "At least one config file is required to set video and data sources"
        )

    config_dict = merge_config_files(config_file_paths)
    render_config = RenderConfig()

    if "target_file_path" in config_dict:
        render_config.target_file_path = config_dict["target_file_path"]
    if "fourcc_codec" in config_dict:
        render_config.fourcc_codec = config_dict["fourcc_codec"]

    if "plotting_method" in config_dict:
        render_config.plotting_method = get_plotting_method(config_dict)
    if "stacking_direction" in config_dict:
        render_config.stacking_direction = get_stacking_direction(config_dict)

    config = Config()
    config.render_config = render_config

    if "video" in config_dict:
        config.video_config = get_video_config(config_dict)
    else:
        raise ValueError(
            "Missing video parameter. "
            "You need to supply a video to render the graphs next to."
        )

    if "sensor" in config_dict:
        config.sensor_configs = get_sensor_configs(config_dict)
    else:
        raise ValueError(
            "Missing sensor parameters. "
            "You need to supply at least one sensor to render next to the video."
        )

    return config


def deep_update(source, overrides):
    """
    Update a nested dictionary or similar mapping.
    Modify ``source`` in place.

    From Nate Glenn, user2709610, charlax, surjikal

    https://stackoverflow.com/a/18394648
    https://stackoverflow.com/a/30655448
    """
    for key, value in overrides.items():
        if isinstance(value, dict) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        elif isinstance(value, list):
            source[key] = source.get(key, []) + value
        else:
            source[key] = overrides[key]
    return source


def get_video_config(cfg: Dict) -> VideoConfig:
    try:
        cfg = cfg["video"][0]
        cfg_type = cfg["type"]
        video_path = cfg["path"]
        del cfg["path"]
        del cfg["type"]

        if cfg_type == "imu":
            result = CameraImuVideoConfig(**cfg)
        elif cfg_type == "kinect":
            result = KinectVideoConfig(**cfg)
        elif cfg_type == "videoImuApp":
            result = VideoImuCaptureAppVideoConfig(**cfg)
        elif cfg_type == "raw":
            result = RawVideoConfig()
        elif cfg_type == "openpose":
            result = OpenPoseVideoConfig()
        else:
            raise ValueError(f"Unknown video config type {cfg_type}")

        result.path = video_path
    except KeyError as e:
        raise KeyError(f"Missing key '{e.args[0]}' in video config: {pformat(cfg)}")
    return result


def get_sensor_configs(config_dict: Dict) -> Dict[str, SensorConfig]:
    result_dict = {}
    i, cfg = None, None
    try:
        config_dict = config_dict["sensor"]
        for i, cfg in enumerate(config_dict):
            cfg_type = cfg["type"]
            cleaned_cfg = cfg.copy()
            del cleaned_cfg["path"]
            del cleaned_cfg["type"]

            if cfg_type == "manually-synced":
                result = ManuallySynchronizedSensorConfig(
                    offset_seconds=cfg.get("offset_seconds", 0.0)
                )
            elif cfg_type == "camera-imu-synced":
                result = ImuSynchronizedSensorConfig(**cleaned_cfg)
            elif cfg_type == "joint-synced":
                result = JointSynchronizedSensorConfig(**cleaned_cfg)
            else:
                raise ValueError(f"Unknown sensor config type {cfg_type}")

            if "start_time" in cfg:
                result.start_time = pd.to_datetime(cfg["start_time"])
            if "end_time" in cfg:
                result.end_time = pd.to_datetime(cfg["end_time"])
            if "graph_groups" in cfg:
                result.graph_groups = cfg["graph_groups"]

            if "name" in cfg:
                result.name = cfg["name"]
            result.path = cfg["path"]
            result_dict[str(i)] = result
    except KeyError as e:
        raise KeyError(
            f"Missing key '{e.args[0]}' in sensor config {i}: {pformat(cfg)}"
        )

    return result_dict


def get_plotting_method(config_dict: Dict) -> PlottingMethod:
    config_plotting_method = config_dict.get("plotting_method", "N/A")
    try:
        return PlottingMethod[config_plotting_method.upper()]
    except KeyError:
        raise KeyError(f"Could not parse plotting_method {config_plotting_method}")


def get_stacking_direction(config_dict: Dict) -> StackingDirection:
    config_stacking_direction = config_dict.get("stacking_direction", "N/A")
    try:
        return StackingDirection[config_stacking_direction.upper()]
    except KeyError:
        raise KeyError(
            f"Could not parse stacking_direction {config_stacking_direction}"
        )



from datetime import datetime

import pandas as pd
import pytest

from sevivi.config import (
    RenderConfig,
    SensorConfig,
    ImuSynchronizedSensorConfig,
    ManuallySynchronizedSensorConfig,
    JointSynchronizedSensorConfig,
    find_matching_columns,
    get_graph_groups,
    PlottingMethod,
)
from sevivi.image_provider import GraphImageProvider


def test_graph_count():
    dti = pd.to_datetime([datetime(2018, 1, 3)])
    df = pd.DataFrame(data={"A_x": [1], "A_y": [1], "A_z": [1], "G_x": [1]}, index=dti)
    assert (
        GraphImageProvider(df, SensorConfig(graph_groups=["A_"])).get_graph_count() == 1
    )
    assert (
        GraphImageProvider(
            df, SensorConfig(graph_groups=["A_", "G_"])
        ).get_graph_count()
        == 2
    )
    assert (
        GraphImageProvider(df, SensorConfig(graph_groups=["_"])).get_graph_count() == 1
    )


def test_get_sync_dataframe_cam_imu():
    dti = pd.to_datetime([datetime(2018, 1, 3)])
    df = pd.DataFrame(data={"A_x": [1], "A_y": [1], "A_z": [1], "G_x": [1]}, index=dti)
    gip = GraphImageProvider(df, SensorConfig())

    gip.sensor_config = ImuSynchronizedSensorConfig(sensor_sync_column_selection="A_")
    assert gip.get_sync_dataframe().columns.tolist() == ["A_x", "A_y", "A_z"]

    gip.sensor_config = ImuSynchronizedSensorConfig(
        sensor_sync_column_selection=["A_x"]
    )
    assert gip.get_sync_dataframe().columns.tolist() == ["A_x"]

    gip.sensor_config = ImuSynchronizedSensorConfig(
        sensor_sync_column_selection=["N/A"]
    )
    with pytest.raises(ValueError):
        gip.get_sync_dataframe()


def test_get_sync_dataframe_joints():
    dti = pd.to_datetime([datetime(2018, 1, 3)])
    df = pd.DataFrame(data={"A_x": [1], "A_y": [1], "A_z": [1], "G_x": [1]}, index=dti)
    gip = GraphImageProvider(df, SensorConfig())

    gip.sensor_config = JointSynchronizedSensorConfig(sensor_sync_column_selection="A_")
    assert gip.get_sync_dataframe().columns.tolist() == ["A_x", "A_y", "A_z"]

    gip.sensor_config = JointSynchronizedSensorConfig(
        sensor_sync_column_selection=["A_x"]
    )
    assert gip.get_sync_dataframe().columns.tolist() == ["A_x"]

    gip.sensor_config = JointSynchronizedSensorConfig(
        sensor_sync_column_selection=["N/A"]
    )
    with pytest.raises(ValueError):
        gip.get_sync_dataframe()


def test_get_sync_data_frame_manual():
    dti = pd.to_datetime(
        [datetime(2018, 1, 1), datetime(2018, 1, 2), datetime(2018, 1, 3)]
    )
    df = pd.DataFrame(data={"A": [1, 2, 3]}, index=dti)
    gip = GraphImageProvider(df, ManuallySynchronizedSensorConfig())
    assert gip.get_sync_dataframe() is None


def test_render_graph_axes():
    dti = pd.to_datetime(
        [datetime(2018, 1, 3), datetime(2018, 1, 2), datetime(2018, 1, 1)]
    )
    df = pd.DataFrame(data={"A": [1, 2, 3]}, index=dti)
    graph_image_provider = GraphImageProvider(df, SensorConfig())
    graph_image_provider.render_graph_axes(None, None)


def test_set_offset_positive():
    dti = pd.to_datetime([datetime(2018, 1, 2), datetime(2018, 1, 3)])
    df = pd.DataFrame(data={"A": [1, 3]}, index=dti)
    graph_image_provider = GraphImageProvider(df, ManuallySynchronizedSensorConfig())
    graph_image_provider.set_offset(pd.Timedelta(days=1))
    new_index = graph_image_provider._data.index
    expected = pd.to_datetime([datetime(1970, 1, 1), datetime(1970, 1, 2)])
    # noinspection PyUnresolvedReferences
    assert (new_index == expected).all()


def test_set_offset_negative():
    dti = pd.to_datetime([datetime(2018, 1, 2), datetime(2018, 1, 3)])
    df = pd.DataFrame(data={"A": [1, 3]}, index=dti)
    graph_image_provider = GraphImageProvider(df, ManuallySynchronizedSensorConfig())
    graph_image_provider.set_offset(pd.Timedelta(days=-1))
    new_index = graph_image_provider._data.index
    expected = pd.to_datetime([datetime(1970, 1, 1), datetime(1970, 1, 2)])
    # noinspection PyUnresolvedReferences
    assert (new_index == expected).all()


def test_get_sync_data_frame_bad_index():
    with pytest.raises(ValueError):
        GraphImageProvider(pd.DataFrame(), SensorConfig())


def test_get_graph_groups():
    df = pd.DataFrame(
        data={"A_x": [1], "A_y": [1], "A_z": [1], "G_x": [1]},
        index=pd.to_datetime([datetime(2018, 1, 3)]),
    )
    assert get_graph_groups(df, ["A_"]) == {"A_": ["A_x", "A_y", "A_z"]}
    assert get_graph_groups(df, ["A_", "G_"]) == {
        "A_": ["A_x", "A_y", "A_z"],
        "G_": ["G_x"],
    }
    assert get_graph_groups(df, ["_"]) == {"_": ["A_x", "A_y", "A_z", "G_x"]}
    assert get_graph_groups(df, ["A_x"]) == {"A_x": ["A_x"]}


def test_find_matching_columns():
    df = pd.DataFrame(
        data={"A_x": [1], "A_y": [1], "A_z": [1], "G_x": [1]},
        index=pd.to_datetime([datetime(2018, 1, 3)]),
    )
    assert find_matching_columns(df, "A_") == ["A_x", "A_y", "A_z"]
    assert find_matching_columns(df, ["A_x"]) == ["A_x"]

    with pytest.raises(ValueError):
        find_matching_columns(df, ["N/A"])



from dataclasses import dataclass

from .stacking_direction import StackingDirection
from .plotting_method import PlottingMethod


@dataclass
class RenderConfig:
    """Contains configuration relevant for rendering"""

    stacking_direction: StackingDirection = StackingDirection.HORIZONTAL
    """Should the plots be next to or below the input video?"""
    plotting_method: PlottingMethod = PlottingMethod.MOVING_VERTICAL_LINE
    """The plotting method to use, i.e., how should changes in time be shown?"""
    target_file_path: str = "sevivi.avi"
    """Path where the resulting video file should be stored"""
    fourcc_codec: str = "MJPG"
    """Fourcc (check OpenCV docs) codec of the resulting video"""
    plot_column_count = 2
    """Number of columns to use in the plots. Should be divisible by 2 for StackingDirection.HORIZONTAL"""



"""Sets up the logging format"""
import logging

FORMAT = "[%(lineno)3s - %(funcName)20s() ] %(message)s"
logging.basicConfig(format=FORMAT)
logger = logging.getLogger("sevivi")
logger.setLevel(logging.DEBUG)



from pprint import pformat
from typing import List, Optional

import pandas as pd
from matplotlib.axes import SubplotBase, Axes
from matplotlib.axis import Axis
from matplotlib.figure import Figure

from sevivi.config import (
    RenderConfig,
    find_matching_columns,
    SensorConfig,
    ManuallySynchronizedSensorConfig,
    JointSynchronizedSensorConfig,
    ImuSynchronizedSensorConfig,
    get_graph_groups,
    PlottingMethod,
)
from sevivi.log import logger
from .utils import epochize_index, calculate_index, RenderAxis

logger = logger.getChild("graph_provider")


class GraphImageProvider:
    """
    A GraphImageProvider has some data with a DatetimeIndex to display.
    An offset to that data can be set to synchronize the data with the video.
    Groups of axes that should be rendered together can be set in the config
    """

    group_line_colors = [f"C{i}" for i in range(10)]
    """
    Provide the default matplotlib color cycle for 10 dimensions.
    Overwrite with a bigger color cycle if you have more dimensions
    in a single graph, or want different colors.
    """

    @property
    def name(self) -> str:
        return self.sensor_config.name

    def __init__(
        self,
        data: pd.DataFrame,
        sensor_config: SensorConfig,
        plotting_method: PlottingMethod = PlottingMethod.MOVING_VERTICAL_LINE,
    ):
        self._data = epochize_index(data)
        self.plotting_method = plotting_method
        self.sensor_config = sensor_config

        self.__axs: List[RenderAxis] = []

        self._graph_groups = get_graph_groups(data, sensor_config.graph_groups)
        logger.debug(
            f"Got groups {pformat(self._graph_groups)} "
            f"for columns {list(data.columns)} "
            f"and graph_groups {sensor_config.graph_groups}"
        )

    def set_axs(self, figure: Figure, axes: List[Axes]):
        """Assign the axes that this graph provider may draw to."""
        for axis_idx, (title, cols) in enumerate(self._graph_groups.items()):
            ax = axes[axis_idx]
            ax.set_xlim(self._data.index[0], self._data.index[-1])
            ax.set_ylim(self._data[cols].min().min(), self._data[cols].max().max())

            if self.sensor_config.name != "":
                title = f"{self.sensor_config.name}: {title}"
            ax.set_title(title)

            render_axis = RenderAxis(ax, cols, title)
            for col_idx, col in enumerate(cols):
                color = self.group_line_colors[col_idx]
                line = ax.plot(
                    self._data.index,
                    self._data[col],
                    color=color,
                )[0]

                # draw to cache renderers
                figure.canvas.draw()

                bbox_copy = figure.canvas.copy_from_bbox(line.clipbox)
                render_axis.bounding_boxes[col] = bbox_copy

            render_axis.vline = ax.axvline(0, color="grey")

            self.__axs.append(render_axis)

    def get_sync_dataframe(self) -> Optional[pd.DataFrame]:
        """
        Return a dataframe that can be used to synchronize this GraphDataProvider to a video.
        When an offset between the video data and the data from this graph provider has been
        calculated, the set_offset function can be used to apply the offset.
        """
        is_imu_cam = isinstance(self.sensor_config, ImuSynchronizedSensorConfig)
        is_joint_cam = isinstance(self.sensor_config, JointSynchronizedSensorConfig)
        is_manual_cam = isinstance(self.sensor_config, ManuallySynchronizedSensorConfig)

        if is_manual_cam:
            return None
        elif is_imu_cam or is_joint_cam:
            # noinspection PyUnresolvedReferences
            sync_cols = self.sensor_config.sensor_sync_column_selection
            matching_columns = find_matching_columns(self._data, sync_cols)
            return self._data[matching_columns]
        else:
            raise ValueError("Cannot get sync DF for unknown sensor config type")

    def get_graph_count(self) -> int:
        """
        Get the number of axes this GraphProvider needs to render all its graphs.
        """
        return len(self._graph_groups)

    def render_graph_axes(self, figure: Figure, ts: pd.Timestamp):
        """
        Render to the axes this graph provider has been assigned.
        Figure must be the same figure the axes instances are from.
        Data will be rendered for the given timestamp.
        """
        if self.plotting_method == PlottingMethod.MOVING_VERTICAL_LINE:
            for render_axis in self.__axs:

                for bbox in render_axis.bounding_boxes.values():
                    figure.canvas.restore_region(bbox)

                # noinspection PyTypeChecker
                vline_position = calculate_index(ts, self._data.index)
                # noinspection PyTypeChecker
                render_axis.vline.set_xdata([vline_position, vline_position])
                render_axis.ax.draw_artist(render_axis.vline)
        else:
            raise NotImplementedError("Only VLINE implemented so far")

    def set_offset(self, offset: pd.Timedelta):
        """
        Apply an offset to the data of this graph provider.
        Positive and negative offsets are supported.
        GraphProvider instances for ManuallySynchronizedSensorConfigs ignore the provided
        argument and instead add the manually configured offset.
        """
        if isinstance(self.sensor_config, ManuallySynchronizedSensorConfig):
            offset = pd.to_timedelta(self.sensor_config.offset_seconds, unit="s")

        self._data.index += offset



import os

import pytest


@pytest.fixture(scope="function")
def run_in_repo_root(request):
    if os.getcwd().endswith("tests"):
        os.chdir("..")
    yield
    os.chdir(request.config.invocation_dir)



from typing import TypeVar, Generator, List, Tuple, Iterable

T = TypeVar("T")


def progress_bar(
    iterable: Iterable[T],
    total: int,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="█",
) -> Generator[Tuple[int, T], None, None]:
    """
    Call in a loop to create terminal progress bar. Adapted from https://stackoverflow.com/a/34325723.
    Yields

    :param iterable:    - Required  : Items from this list will be yielded to the for-loop
    :param total:       - Required  : Number of items that will be yielded
    :param prefix:      - Optional  : prefix string (Str)
    :param suffix:      - Optional  : suffix string (Str)
    :param decimals:    - Optional  : positive number of decimals in percent complete (Int)
    :param length:      - Optional  : character length of bar (Int)
    :param fill:        - Optional  : bar fill character (Str)
    """

    # Progress Bar Printing Function
    def print_progress_bar(iteration):
        percent = ("{0:." + str(decimals) + "f}").format(
            100 * (iteration / float(total))
        )
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + "-" * (length - filledLength)
        print(f"\r{prefix} |{bar}| {percent}% {suffix}", end="\r")

    # Initial Call
    print_progress_bar(0)
    # Update Progress Bar
    for i, item in enumerate(iterable):
        yield i, item
        print_progress_bar(i + 1)
    # Print New Line on Complete
    print()



from typing import Dict

import pandas as pd
import pytest
from sevivi.config import config_reader, PlottingMethod, StackingDirection, VideoConfig
from sevivi.config.config_reader import deep_update
from sevivi.config.config_types.sensor_config import (
    SensorConfig,
    ManuallySynchronizedSensorConfig,
    JointSynchronizedSensorConfig,
    ImuSynchronizedSensorConfig,
)
from sevivi.config.config_types.video_config import (
    CameraImuVideoConfig,
    KinectVideoConfig,
    OpenPoseVideoConfig,
    RawVideoConfig,
)


def test_combination(run_in_repo_root):
    config_reader.read_configs(
        (
            "test_files/configs/basic_config.toml",
            "test_files/configs/video_configs/kinect.toml",
            "test_files/configs/sensor_configs/camera_imu_synchronization.toml",
            "test_files/configs/sensor_configs/joint_synchronization.toml",
            "test_files/configs/sensor_configs/manual_synchronization.toml",
        )
    )


def test_zero_config():
    with pytest.raises(ValueError):
        # noinspection PyTypeChecker
        config_reader.read_configs(None)
    with pytest.raises(ValueError):
        config_reader.read_configs(())


def test_missing_config_file():
    with pytest.raises(FileNotFoundError):
        config_reader.merge_config_files(("lnjaerovfgnaeorcin",))
    with pytest.raises(FileNotFoundError):
        config_reader.read_configs(("lnjaerovfgnaeorcin",))


def test_single_config_file(run_in_repo_root):
    assert _conf_dict("basic_config") is not None


def test_multi_config(run_in_repo_root):
    config = _conf_dict("basic_config")
    config_multi = _conf_dict("basic_config", "basic_config")
    assert config == config_multi

    assert (
        config_reader.get_plotting_method(config) == PlottingMethod.MOVING_VERTICAL_LINE
    )
    config_parallel_ingest = _conf_dict("basic_config", "push_in_plotting_method")
    assert (
        config_reader.get_plotting_method(config_parallel_ingest)
        == PlottingMethod.PUSH_IN
    )


def test_plotting_method(run_in_repo_root):
    config = _conf_dict("basic_config")
    assert (
        config_reader.get_plotting_method(config) == PlottingMethod.MOVING_VERTICAL_LINE
    )

    config = _conf_dict("push_in_plotting_method")
    assert config_reader.get_plotting_method(config) == PlottingMethod.PUSH_IN

    with pytest.raises(KeyError):
        config_reader.get_plotting_method(_conf_dict("bad_plotting_method"))


def test_stack_direction(run_in_repo_root):
    config = _conf_dict("basic_config")
    assert config_reader.get_stacking_direction(config) == StackingDirection.HORIZONTAL

    config = _conf_dict("vertical_stacking")
    assert config_reader.get_stacking_direction(config) == StackingDirection.VERTICAL

    with pytest.raises(KeyError):
        config_reader.get_stacking_direction(_conf_dict("bad_stacking"))


def test_read_imu_video_config(run_in_repo_root):
    imu_conf = _vid_conf("imu")
    assert imu_conf.path == "test_files/videos/imu_sync.mp4"
    assert isinstance(imu_conf, CameraImuVideoConfig)
    assert imu_conf.imu_path == "test_files/kinect_imu.csv.gz"


def test_read_kinect_video_config(run_in_repo_root):
    imu_conf = _vid_conf("kinect")
    assert imu_conf.path == "test_files/kinect.mkv"
    assert isinstance(imu_conf, KinectVideoConfig)
    assert imu_conf.skeleton_path_3d == "test_files/kinect.csv.gz"


def test_read_openpose_video_config(run_in_repo_root):
    imu_conf = _vid_conf("openpose")
    assert imu_conf.path == "test_files/openpose.mkv"
    assert isinstance(imu_conf, OpenPoseVideoConfig)


def test_read_raw_video_config(run_in_repo_root):
    imu_conf = _vid_conf("raw")
    assert imu_conf.path == "test_files/raw.mkv"
    assert isinstance(imu_conf, RawVideoConfig)


def test_dual_imu(run_in_repo_root):
    conf = _sensor_conf("dual_imu_base")
    assert len(conf) == 2


def test_base_imu(run_in_repo_root):
    conf = _sensor_conf("single_imu_base")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf, ManuallySynchronizedSensorConfig)
    assert conf.offset_seconds == 0.0


def test_manual_sync_imu(run_in_repo_root):
    conf = _sensor_conf("manual_synchronization")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf, ManuallySynchronizedSensorConfig)
    assert conf.offset_seconds == 123.4


def test_joint_sync_imu(run_in_repo_root):
    conf = _sensor_conf("joint_synchronization")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf, JointSynchronizedSensorConfig)
    assert conf.sensor_sync_column_selection == ["AccX", "Accel Y", "ACC Z"]
    assert conf.camera_joint_sync_column_selection == ["AccX", "Accel Y", "ACC Z"]


def test_camera_sync_imu(run_in_repo_root):
    conf = _sensor_conf("camera_imu_synchronization")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf, ImuSynchronizedSensorConfig)
    assert conf.sensor_sync_column_selection == ["AccX", "Accel Y"]
    assert conf.path == "test_files/sensor_imu.csv.gz"


def test_camera_sync_imu_single_col(run_in_repo_root):
    conf = _sensor_conf("camera_imu_synchronization_single_col")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf, ImuSynchronizedSensorConfig)
    assert conf.sensor_sync_column_selection == "AccX"
    assert conf.path == "test_files/camera_imu.csv.gz"


def test_imu_merging(run_in_repo_root):
    conf = _sensor_conf("single_imu_base", "single_imu_base")
    assert len(conf) == 2
    assert conf["0"] == conf["1"]


def test_sensor_with_end_time(run_in_repo_root):
    conf = _sensor_conf("end_time")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf.end_time, pd.Timestamp)


def test_sensor_with_start_time(run_in_repo_root):
    conf = _sensor_conf("start_time")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf.start_time, pd.Timestamp)


def test_sensor_with_start_and_end_time(run_in_repo_root):
    conf = _sensor_conf("start_end_time")
    assert len(conf) == 1
    conf = conf["0"]
    assert isinstance(conf.start_time, pd.Timestamp)
    assert isinstance(conf.end_time, pd.Timestamp)


def test_sensor_missing_attributes(run_in_repo_root):
    with pytest.raises(KeyError):
        _sensor_conf("missing_attributes")


def test_video_missing_attributes(run_in_repo_root):
    with pytest.raises(KeyError):
        _vid_conf("missing_attributes")


def _vid_conf(*names: str) -> VideoConfig:
    files = [f"test_files/configs/video_configs/{name}.toml" for name in names]
    return config_reader.get_video_config(
        config_reader.merge_config_files(tuple(files))
    )


def _sensor_conf(*names: str) -> Dict[str, SensorConfig]:
    files = [f"test_files/configs/sensor_configs/{name}.toml" for name in names]
    return config_reader.get_sensor_configs(
        config_reader.merge_config_files(tuple(files))
    )


def _conf_dict(*names: str) -> Dict:
    files = [f"test_files/configs/{name}.toml" for name in names]
    return config_reader.merge_config_files(tuple(files))


def test_deep_update():
    """
    Mostly from charlax, surjikal

    https://stackoverflow.com/a/18394648
    https://stackoverflow.com/a/30655448
    """
    source = {"hello1": 1}
    overrides = {"hello2": 2}
    deep_update(source, overrides)
    assert source == {"hello1": 1, "hello2": 2}

    source = {"hello": "to_override"}
    overrides = {"hello": "over"}
    deep_update(source, overrides)
    assert source == {"hello": "over"}

    source = {"hello": {"value": "to_override", "no_change": 1}}
    overrides = {"hello": {"value": "over"}}
    deep_update(source, overrides)
    assert source == {"hello": {"value": "over", "no_change": 1}}

    source = {"hello": {"value": "to_override", "no_change": 1}}
    overrides = {"hello": {"value": {}}}
    deep_update(source, overrides)
    assert source == {"hello": {"value": {}, "no_change": 1}}

    source = {"hello": {"value": {}, "no_change": 1}}
    overrides = {"hello": {"value": 2}}
    deep_update(source, overrides)
    assert source == {"hello": {"value": 2, "no_change": 1}}


def test_multi_video_config(run_in_repo_root):
    config_reader.read_configs(
        (
            "test_files/configs/basic_config.toml",
            "test_files/configs/video_configs/kinect.toml",
            "test_files/configs/video_configs/kinect.toml",
            "test_files/configs/sensor_configs/camera_imu_synchronization.toml",
        )
    )


def test_missing_video_config(run_in_repo_root):
    with pytest.raises(ValueError):
        config_reader.read_configs(
            (
                "test_files/configs/basic_config.toml",
                "test_files/configs/sensor_configs/camera_imu_synchronization.toml",
            )
        )


def test_missing_sensor_config(run_in_repo_root):
    with pytest.raises(ValueError):
        config_reader.read_configs(
            (
                "test_files/configs/basic_config.toml",
                "test_files/configs/video_configs/kinect.toml",
            )
        )


def test_default_target_file(run_in_repo_root):
    cfg = config_reader.read_configs(
        (
            "test_files/configs/basic_config.toml",
            "test_files/configs/video_configs/kinect.toml",
            "test_files/configs/sensor_configs/camera_imu_synchronization.toml",
        )
    )
    cfg.render_config.target_file_path = "sevivi.mp4"


def test_modified_target_file(run_in_repo_root):
    cfg = config_reader.read_configs(
        (
            "test_files/configs/basic_config.toml",
            "test_files/configs/video_configs/kinect.toml",
            "test_files/configs/sensor_configs/camera_imu_synchronization.toml",
            "test_files/configs/change_target_file.toml",
        )
    )
    cfg.render_config.target_file_path = "test.mp4"



from sevivi.config.config_reader import read_configs


def test_get_missing_files_all_available(run_in_repo_root):
    config = read_configs(
        ("test_files/test-data-configs/kinect_sync_3d_hand_movement.toml",)
    )
    assert len(config.get_missing_files()) == 0


def test_get_missing_files_four_missing(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/missing_files.toml",))
    missing = config.get_missing_files()
    assert len(missing) == 4
    assert "test_files/videos/todo.mp4" in missing
    assert "test_files/todo0.csv.gz" in missing
    assert "test_files/todo1.csv.gz" in missing
    assert "test_files/todo2.csv.gz" in missing



from scipy import signal
from scipy import interpolate

import numpy as np
import pandas as pd


def normalize_signal(data: np.ndarray) -> np.ndarray:
    """Normalize a signal to have zero mean and unit variance"""
    return (data - data.mean()) / data.std()


def resample_data(
    data: np.ndarray, current_sampling_rate: int, new_sampling_rate: int, mode="cubic"
) -> np.ndarray:
    """Resamples a signal to the desired new sampling frequency"""
    frames, features = data.shape
    x = np.arange(len(data)) / current_sampling_rate
    num = int(x[-1] * new_sampling_rate)  # Define new constant sampling points
    xx = np.linspace(x[0], x[-1], num)

    result = []
    for feature in range(features):
        y = data[:, feature]
        f = interpolate.interp1d(x, y, kind=mode)
        result.append(f(xx))

    return np.array(result).T


def calculate_magnitude(data: np.ndarray) -> np.ndarray:
    """Calculates the magnitude for given (x,y,z) axes stored in numpy array"""
    assert data.shape[1] == 3, f"Numpy array should have 3 axes, got {data.shape[1]}"
    return np.sqrt(np.square(data).sum(axis=1))


def calculate_offset_in_seconds_using_cross_correlation(
    ref_signal: np.ndarray, target_signal: np.ndarray, sampling_frequency: int
) -> float:
    """Calculates the temporal offset between the two given numpa arrays"""
    corr = signal.correlate(ref_signal, target_signal)
    shift_in_samples = np.argmax(corr) - len(target_signal) - 1
    return shift_in_samples / sampling_frequency


def calculate_sampling_frequency_from_timestamps(timestamps: pd.DatetimeIndex) -> int:
    """Calculates the sampling frequency from the timestamp column in dataframe"""
    ts_data_ns = timestamps.astype(np.uint64).to_numpy()
    deltas = np.diff(ts_data_ns) / 1e9
    sampling_freq = 1 / deltas.mean()
    return round(sampling_freq)



from .azure_image_provider import AzureProvider
from .plain_video_image_provider import PlainVideoImageProvider
from .imu_camera_image_provider import ImuCameraImageProvider
from .video_provider import VideoImageProvider
from .video_imu_capture_app.video_imu_capture_app_image_provider import (
    VideoImuCaptureAppImageProvider,
)



from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any

import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from sevivi.log import logger

logger = logger.getChild("graph_provider")


def calculate_index(
    target_ts: pd.Timestamp, timestamps: pd.DatetimeIndex
) -> pd.Timestamp:
    """
    Return the first index value after the target timestamp if the exact timestamp is not available
    """
    # noinspection PyUnresolvedReferences
    target_beyond_available = (target_ts > timestamps).all()

    if target_beyond_available:
        return timestamps[-1]
    elif target_ts in timestamps:
        return target_ts
    else:
        return timestamps[timestamps > target_ts][0]


def epochize_index(input: pd.DataFrame) -> pd.DataFrame:
    """
    Put the first index value of the input dataframe exactly on the unix epoch.
    """
    if not isinstance(input.index, pd.DatetimeIndex):
        raise ValueError("All graph data must use a DatetimeIndex!")

    zeroed_index = input.index.view(np.int64)
    zeroed_index -= zeroed_index[0]
    input.index = pd.to_datetime(zeroed_index, unit="ns")
    return input


@dataclass
class RenderAxis:
    """Encapsulate all information about the axes of a GraphProvider needed for efficient drawing."""

    ax: Axes = None
    """The matplotlib axis to render to"""
    columns: List[str] = None
    """The columns of the dataframe plotted in this axis"""
    name: str = None
    """Human-readable title prefix of this axis"""
    vline: Optional[Line2D] = None
    """Vline artist"""
    bounding_boxes: Dict[str, Any] = field(default_factory=dict)
    """Maps all columns in this axis to a BufferRegion of the graphic inside their bounding box"""



from dataclasses import dataclass
from typing import Generator


@dataclass
class Dimensions:
    """Used to store width and height of an image in a single object"""

    w: int
    """width in pixels"""
    h: int
    """height in pixels"""

    def __iter__(self) -> Generator[int, None, None]:
        """
        Iterate over width, height.
        Can be used to convert this into a (width, height) tuple by calling
        tuple(dimension)
        """
        yield self.w
        yield self.h



import os
from dataclasses import dataclass
from typing import Optional, Union, List

import pandas as pd


@dataclass
class SensorConfig:
    """
    SensorConfig instances allow specifying the data taken and synchronized from sensors.
    """

    path: str = None
    """Path to the sensor data csv"""
    name: str = ""
    """Human readable name of the sensor for display purposes"""
    start_time: Optional[pd.Timestamp] = None
    """Start time of the sensor data. This is useful to select a portion of data from a longer recording"""
    end_time: Optional[pd.Timestamp] = None
    """End time of the sensor data. This is useful to select a portion of data from a longer recording"""
    graph_groups: Optional[List[str]] = None
    """List of in/pre/postfixes that select a number of columns to be graphed in the same axis."""

    def get_missing_files(self) -> List[str]:
        """Returns a list of all missing files for this config"""
        if not os.path.isfile(self.path):
            return [self.path]
        return []


@dataclass
class ManuallySynchronizedSensorConfig(SensorConfig):
    """
    This SensorConfig provides the information necessary to synchronize a sensor manually
    """

    offset_seconds: float = 0.0
    """Offset the select data chunk, or the entire graph data, by this number of seconds"""


@dataclass
class JointSynchronizedSensorConfig(SensorConfig):
    """
    This SensorConfig provides the information necessary to synchronize by using a number of columns from the sensor data
    to align to axes from a specific joint.

    :Example:
    path: /tmp/my_ankle_imu.csv.gz
    start_time: 00:00:00.000000
    sensor_sync_column_selection: ["AX", "accel Y", "ACCELERATION_Z"]
    camera_joint_sync_column_selection: ACCELERATION_MAG
    """

    camera_joint_sync_column_selection: Union[str, List[str]] = None
    """Source columns from the camera skeleton to synchronize to"""
    sensor_sync_column_selection: Union[str, List[str]] = None
    """Columns from the sensor data to synchronize to"""


@dataclass
class ImuSynchronizedSensorConfig(SensorConfig):
    """This SensorConfig provides the information necessary to synchronize data to a video by camera IMU."""

    sensor_sync_column_selection: Union[str, List[str]] = None
    """
    Name of the column or list of the columns that contain the IMU data to synchronize with the camera IMU.
    This allows to e.g., select the acceleration columns from your sensor. The names of these columns coming from the
    camera IMU are configured using CameraImuVideoConfig.
    """

    camera_imu_sync_column_selection: Union[str, List[str]] = None
    """Source columns from the camera IMU to synchronize to"""



"""VideoImageProviders provide the video aroud which the data graphs are rendered"""

from typing import Optional, List, Generator, Tuple

import numpy as np
import pandas as pd

from ..dimensions import Dimensions


class VideoImageProvider:
    """VideoImageProviders provide the video aroud which the data graphs are rendered"""

    def images(self) -> Generator[Tuple[pd.Timestamp, np.ndarray], None, None]:
        """Generate the images to be shown together with their timestamps"""
        raise NotImplementedError(
            "images must be implemented by VideoImageProvider subclasses!"
        )

    def get_sync_dataframe(self, column_names: List[str]) -> Optional[pd.DataFrame]:
        """
        Get a dataframe that can be used to synchronize graph data against this video.
        The given column names can be provided by GraphImageProviders to select specific
        synchronization targets, e.g., the acceleration of a specific joint.
        """
        raise NotImplementedError(
            "get_sync_dataframe must be implemented by VideoImageProvider subclasses!"
        )

    def get_image_count(self) -> int:
        """Get the number of images that will be rendered"""
        raise NotImplementedError(
            "get_image_count must be implemented by VideoImageProvider subclasses!"
        )

    def get_dimensions(self) -> Dimensions:
        """Get the dimensions of the source video in pixels."""
        raise NotImplementedError(
            "get_dimensions must be implemented by VideoImageProvider subclasses!"
        )



from .video_provider import (
    AzureProvider,
    PlainVideoImageProvider,
    ImuCameraImageProvider,
    VideoImageProvider,
    VideoImuCaptureAppImageProvider,
)
from .graph_provider import GraphImageProvider
from .dimensions import Dimensions



from dataclasses import dataclass, field
from typing import List, Dict

from .video_config import VideoConfig
from .render_config import RenderConfig
from .sensor_config import SensorConfig


@dataclass
class Config:
    """Typed configuration for the CLI. Contains render configuration as well as video and sensor config."""

    video_config: VideoConfig = None
    sensor_configs: Dict[str, SensorConfig] = field(default_factory=dict)
    render_config: RenderConfig = None

    def get_missing_files(self) -> List[str]:
        """
        Returns a list of missing files in this config.
        If no files are missing, the list is empty
        """
        missing_files = self.video_config.get_missing_files()
        for sensor_config in self.sensor_configs.values():
            missing_files.extend(sensor_config.get_missing_files())
        return missing_files



from sevivi import video_renderer_from_csv_files, read_configs

from sys import platform as sys_pf
import matplotlib

if sys_pf == "darwin":
    matplotlib.use("TkAgg")


def test_plain_video_render(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/manual_sync.toml",))
    video_renderer = video_renderer_from_csv_files(config)
    video_renderer.render_video()


def test_azure_video_render(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/kinect_sync_walking.toml",))
    video_renderer = video_renderer_from_csv_files(config)
    video_renderer.render_video()


def test_video_imu_capture_app_render(run_in_repo_root):
    config = read_configs(
        ("test_files/test-data-configs/video_imu_capture_app_sync.toml",)
    )
    video_renderer = video_renderer_from_csv_files(config)
    video_renderer.render_video()


def test_camera_imu_render(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/imu_sync.toml",))
    video_renderer = video_renderer_from_csv_files(config)
    video_renderer.render_video()



"""The IMU video image provider provides images from a camera with an integrated IMU together with the IMU data."""
from typing import List, Generator, Tuple

import cv2
import pandas as pd
from read_protobuf import ProtobufReader

from sevivi.image_provider.video_provider.video_imu_capture_app import recording_pb2
from sevivi.log import logger
from sevivi.image_provider.video_provider.video_provider import VideoImageProvider
from sevivi.image_provider.dimensions import Dimensions
from sevivi.config import find_matching_columns

logger = logger.getChild("plain_video_image_provider")


class VideoImuCaptureAppImageProvider(VideoImageProvider):
    """
    The VideoImuCaptureAppImageProvider directly supports the VideoIMUCapture-Android app available at
    https://github.com/DavidGillsjo/VideoIMUCapture-Android/. You can use this app to create a video file with
    associated IMU data.

    By shaking both the sensors and the smartphone before and after you record data, you can introduce markers for
    the correlation-based synchronization to work. If synchronization doesn't work, you should provide a custom
    synchronizer implementation.
    """

    def __init__(self, video_path: str, imu_pb_path: str):
        self.__video = cv2.VideoCapture(video_path)

        try:
            with open(imu_pb_path, "rb") as f:
                message_data = f.read()
            message = recording_pb2.VideoCaptureData.FromString(message_data)
            data = ProtobufReader().interpret_message(message)
        except Exception as e:
            raise RuntimeError(
                f"Could not read protobuf file {imu_pb_path}. Is it from VideoIMUCapture_v0.12.apk?"
            ) from e

        timestamps = [item["time_ns"] for item in data["imu"]]
        imu_accel_data = [item["accel"] for item in data["imu"]]
        self.data: pd.DataFrame = pd.DataFrame(
            imu_accel_data,
            index=timestamps,
            columns=["ACCELERATION_X", "ACCELERATION_Y", "ACCELERATION_Z"],
        )

        # noinspection PyTypeChecker
        self.data.index = pd.to_datetime(self.data.index, unit="ns")

    def get_sync_dataframe(self, column_names: List[str]) -> pd.DataFrame:
        """
        Returns the selected portion of the IMU data.
        """
        matching_columns = find_matching_columns(self.data, column_names)
        return self.data[matching_columns]

    def images(self) -> Generator[Tuple[pd.Timestamp, bytes], None, None]:
        """Generate the images to be shown together with their timestamps"""
        while self.__video.isOpened():
            frame_exists, frame = self.__video.read()
            if frame_exists:
                ts = self.__video.get(cv2.CAP_PROP_POS_MSEC)
                yield pd.to_datetime(ts, unit="ms"), frame
            else:
                break

    def get_image_count(self) -> int:
        """Get the number of images that will be rendered"""
        return int(self.__video.get(cv2.CAP_PROP_FRAME_COUNT))

    def get_dimensions(self) -> Dimensions:
        """Get the dimensions of the source video in pixels."""
        return Dimensions(
            w=int(self.__video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            h=int(self.__video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )



import sys
from argparse import ArgumentParser
from pprint import pformat
from typing import List

from sevivi.config import Config, read_configs
from sevivi.video_renderer import video_renderer_from_csv_files
from sevivi.log import logger


def parse_arguments(args: List[str]) -> Config:
    """Parse the config file locations from input args and return the specified configuration"""
    parser = ArgumentParser()
    parser.add_argument(
        "--output",
        dest="target_file_path",
        type=str,
        help="Set the output file. Overwrites config value.",
        required=False,
    )
    parser.add_argument(
        "config",
        nargs="+",
        help="Configuration files. Later files overwrite earlier ones. Only the last video section is used. "
        "All given sensor configs are interpreted as a list, rather than overwriting earlier configuration.",
    )
    args = parser.parse_args(args)
    result = read_configs(args.config)

    if "target_file_path" in args and args.target_file_path is not None:
        result.render_config.target_file_path = args.target_file_path

    return result


def run():
    config = parse_arguments(sys.argv[1:])
    logger.debug(f"Sevivi config is: {pformat(config)}")
    logger.info(f"Rendering video to {config.render_config.target_file_path}")
    video_renderer = video_renderer_from_csv_files(config)
    video_renderer.render_video()



def test_manually_creating_video_renderer(run_in_repo_root):
    import pandas as pd

    from sevivi.config import RenderConfig, ManuallySynchronizedSensorConfig
    from sevivi.image_provider import (
        GraphImageProvider,
        VideoImuCaptureAppImageProvider,
    )
    from sevivi.video_renderer import VideoRenderer

    video_provider = VideoImuCaptureAppImageProvider(
        video_path="test_files/videos/imu_sync.mp4",
        imu_pb_path="test_files/sensors/video_imu_capture_app/video_meta.pb3",
    )

    # create a GraphImageProvider for each of your sensors
    sensor_config = ManuallySynchronizedSensorConfig()
    sensor_config.offset_seconds = 0.0
    sensor_config.name = "Human-Readable Name"
    sensor_config.path = "test_files/sensors/imu_synchronization/camera_imu.csv.gz"
    data = pd.read_csv(sensor_config.path, index_col=0, parse_dates=True)
    graph_image_provider = GraphImageProvider(data, sensor_config)

    # render the video
    renderer = VideoRenderer(RenderConfig(), video_provider, [graph_image_provider])
    renderer.render_video()



from enum import Enum


class StackingDirection(Enum):
    """How to stack plots and source video, i.e., to place graphs below or right and left of the source video."""

    HORIZONTAL = 0
    """Place graphs right and left of the source video"""
    VERTICAL = 1
    """Place graphs below source video."""



import pytest

from sevivi.config import Config
from sevivi.main import parse_arguments


def test_parse_arguments_empty():
    with pytest.raises(SystemExit):
        parse_arguments([])


def test_parse_arguments_missing_config_file():
    with pytest.raises(FileNotFoundError):
        parse_arguments(["amrnvo;eurnvf"])


def test_parse_arguments_single(run_in_repo_root):
    isinstance(parse_arguments(["test_files/test-data-configs/imu_sync.toml"]), Config)


def test_parse_arguments_multiple(run_in_repo_root):
    isinstance(
        parse_arguments(
            [
                "test_files/test-data-configs/imu_sync.toml",
                "test_files/configs/vertical_stacking.toml",
            ]
        ),
        Config,
    )



from math import ceil
from typing import Tuple, Optional, List, Callable

import cv2
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from sevivi.config import (
    RenderConfig,
    JointSynchronizedSensorConfig,
    ImuSynchronizedSensorConfig,
    ManuallySynchronizedSensorConfig,
    StackingDirection,
)
from sevivi.image_provider import GraphImageProvider, VideoImageProvider, Dimensions
from sevivi.log import logger
from sevivi.synchronizer.synchronizer import get_synchronization_offset
from .progress_bar import progress_bar

logger = logger.getChild("video_renderer")

DPI = 100


class VideoRenderer:
    """
    Renders a number of graphs by the GraphImageProviders next to a video from the
    VideoImageProvider as specified by the given RenderConfig
    """

    def __init__(
        self,
        render_config: RenderConfig,
        video_provider: VideoImageProvider,
        graph_providers: List[GraphImageProvider],
        synchronization_function: Callable[
            [pd.DataFrame, pd.DataFrame, bool, Optional[bool]], pd.Timedelta
        ] = get_synchronization_offset,
    ):
        """
        :param render_config: rendering configuration, e.g., horizontal or vertical stacking
        :param video_provider: the video provider for the video
        :param graph_providers: the graph providers for each sensor
        :param synchronization_function: a synchronization function that adheres to the specifications of
                                         sevivi.synchronizer.synchronizer.get_synchronization_offset
        """
        self.synchronization_function = synchronization_function
        self.render_config = render_config
        self.video_provider = video_provider
        self.graph_providers = graph_providers

        self._graph_count = sum([gp.get_graph_count() for gp in self.graph_providers])
        self.__src_vid_dims = self.video_provider.get_dimensions()
        self._plot_dims, self.__tgt_vid_dims = self._prepare_dimensions()
        self._fig, self._axs = self._prepare_figure()
        self._prepare_graph_providers()

    def _prepare_dimensions(self) -> Tuple[Dimensions, Dimensions]:
        """Calculate the desired plot and target video dimensions"""
        src_vid_dim = self.__src_vid_dims
        if self.render_config.stacking_direction == StackingDirection.VERTICAL:
            plot_w, plot_h = src_vid_dim.w, 200 * self._graph_count
            video_w, video_h = src_vid_dim.w, src_vid_dim.h + plot_h
        else:
            plot_w, plot_h = src_vid_dim.w, src_vid_dim.h
            video_w, video_h = src_vid_dim.w // 2 + plot_w, src_vid_dim.h
        return Dimensions(plot_w, plot_h), Dimensions(video_w, video_h)

    def _prepare_figure(self) -> Tuple[Figure, np.ndarray]:
        """Create a figure with subplots and return the figure and its subplots"""
        graph_rows = ceil(self._graph_count / self.render_config.plot_column_count)
        graph_cols = self.render_config.plot_column_count

        figsize = self._plot_dims.w / DPI, self._plot_dims.h / DPI
        fig, axs = plt.subplots(
            graph_rows,
            graph_cols,
            figsize=figsize,
            squeeze=False,
            dpi=DPI,
        )
        plt.tight_layout()

        return fig, axs.ravel()

    def _prepare_graph_providers(self):
        """Set offsets to graph providers and assigns the available axes to the graph providers"""
        assigned_axis_count = 0
        for gp in self.graph_providers:
            gp.set_offset(self._calc_offset(gp))

            axis_idx_limit = assigned_axis_count + gp.get_graph_count()
            gp.set_axs(self._fig, self._axs[assigned_axis_count:axis_idx_limit])
            logger.debug(
                f"Assigned axis {assigned_axis_count}:{axis_idx_limit} to GP {gp.sensor_config.name}"
            )
            assigned_axis_count += gp.get_graph_count()

    def _calc_offset(
        self, graph_provider: GraphImageProvider
    ) -> Optional[pd.Timedelta]:
        """
        Calculate the offset to a graph provider using the synchronization module.
        Returns None in case of manual synchronization
        """
        config = graph_provider.sensor_config

        if isinstance(config, ManuallySynchronizedSensorConfig):
            return None
        elif isinstance(config, JointSynchronizedSensorConfig):
            use_video_gradient_for_offset = True
            video_sync_cols = config.camera_joint_sync_column_selection
        elif isinstance(config, ImuSynchronizedSensorConfig):
            use_video_gradient_for_offset = False
            video_sync_cols = config.camera_imu_sync_column_selection
        else:
            raise RuntimeError(f"Unknown sensor cfg type: {type(config)}")

        video_sync_df = self.video_provider.get_sync_dataframe(video_sync_cols)
        graph_sync_df = graph_provider.get_sync_dataframe()
        offset = self.synchronization_function(
            video_sync_df, graph_sync_df, use_video_gradient_for_offset
        )

        logger.debug(f"Graph {graph_provider.sensor_config.name} gets offset {offset}")
        return offset

    def stitch_plot_image(self, ts: pd.Timestamp) -> np.ndarray:
        # self.graph_providers["1"].render_graph_axes(self._fig, ts)
        for gp in self.graph_providers:
            gp.render_graph_axes(self._fig, ts)

        canvas_buffer = self._fig.canvas.buffer_rgba()
        return cv2.cvtColor(np.asarray(canvas_buffer), cv2.COLOR_RGBA2BGR)

    def render_video(self):
        """Renders the video from given VideoImageProvider and GraphImageProviders using the given RenderConfig."""
        src_vid_dim = self.__src_vid_dims
        image = np.empty(
            (self.__tgt_vid_dims.h, self.__tgt_vid_dims.w, 3), dtype=np.uint8
        )

        writer = cv2.VideoWriter(
            self.render_config.target_file_path,
            cv2.VideoWriter_fourcc(*self.render_config.fourcc_codec),
            32,
            tuple(self.__tgt_vid_dims),
        )

        video_provider = self.video_provider
        images, image_count = video_provider.images(), video_provider.get_image_count()
        for index, (ts, src_image) in progress_bar(images, image_count):
            plot_image = self.stitch_plot_image(ts)

            if self.render_config.stacking_direction == StackingDirection.VERTICAL:
                # add the original video image
                image[: src_vid_dim.h, :, :] = src_image
                below_src_vid = slice(src_vid_dim.h, src_vid_dim.h + self._plot_dims.h)
                image[below_src_vid, :, :] = plot_image
            else:
                # add the center half of the original video image
                plot_center = self._plot_dims.w // 2
                tgt_vid_left = slice(0, plot_center)
                tgt_vid_center = slice(plot_center, plot_center + src_vid_dim.w // 2)
                tgt_vid_right = slice(src_vid_dim.w // 2 + plot_center, None, None)

                source_video_half = slice(src_vid_dim.w // 4, 3 * src_vid_dim.w // 4)

                image[:, tgt_vid_left, :] = plot_image[:, tgt_vid_left, :]
                image[:, tgt_vid_center, :] = src_image[:, source_video_half, :]
                image[:, tgt_vid_right, :] = plot_image[:, plot_center:, :]

            writer.write(image)
        writer.release()



"""The plain video image provider provides image from any video supported by openCV with manual synchronization."""
from typing import List, Generator, Tuple

import cv2
import pandas as pd

from sevivi.log import logger
from .video_provider import VideoImageProvider
from ..dimensions import Dimensions

logger = logger.getChild("plain_video_image_provider")


class PlainVideoImageProvider(VideoImageProvider):
    """The plain video image provider provides image from any video supported by openCV with manual synchronization."""

    def __init__(self, video_path: str):
        self.__video = cv2.VideoCapture(video_path)

    def get_sync_dataframe(self, column_names: List[str]) -> None:
        """
        Plain video doesn't have any data to synchronize against, so the sync dataframe is None.
        """
        return None

    def images(self) -> Generator[Tuple[pd.Timestamp, bytes], None, None]:
        """Generate the images to be shown together with their timestamps"""
        while self.__video.isOpened():
            frame_exists, frame = self.__video.read()
            if frame_exists:
                ts = self.__video.get(cv2.CAP_PROP_POS_MSEC)
                yield pd.to_datetime(ts, unit="ms"), frame
            else:
                break

    def get_image_count(self) -> int:
        """Get the number of images that will be rendered"""
        return int(self.__video.get(cv2.CAP_PROP_FRAME_COUNT))

    def get_dimensions(self) -> Dimensions:
        """Get the dimensions of the source video in pixels."""
        return Dimensions(
            w=int(self.__video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            h=int(self.__video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )



from google.protobuf import (
    timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2,
)

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: recording.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="recording.proto",
    package="videoimu",
    syntax="proto3",
    serialized_options=b"\n\033se.lth.math.videoimucaptureB\017RecordingProtos",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0frecording.proto\x12\x08videoimu\x1a\x1fgoogle/protobuf/timestamp.proto"\xba\x06\n\nCameraInfo\x12\x18\n\x10intrinsic_params\x18\x01 \x03(\x02\x12\x19\n\x11\x64istortion_params\x18\x02 \x03(\x02\x12#\n\x1boptical_image_stabilization\x18\x03 \x01(\x08\x12\x1b\n\x13video_stabilization\x18\x04 \x01(\x08\x12\x1d\n\x15\x64istortion_correction\x18\n \x01(\x08\x12\x1a\n\x12sensor_orientation\x18\x0e \x01(\x05\x12@\n\x11\x66ocus_calibration\x18\x05 \x01(\x0e\x32%.videoimu.CameraInfo.FocusCalibration\x12>\n\x10timestamp_source\x18\x06 \x01(\x0e\x32$.videoimu.CameraInfo.TimestampSource\x12\x43\n\x13lens_pose_reference\x18\x07 \x01(\x0e\x32&.videoimu.CameraInfo.LensPoseReference\x12\x1a\n\x12lens_pose_rotation\x18\x08 \x03(\x02\x12\x1d\n\x15lens_pose_translation\x18\t \x03(\x02\x12-\n\nresolution\x18\x0b \x01(\x0b\x32\x19.videoimu.CameraInfo.Size\x12\x43\n pre_correction_active_array_size\x18\x0c \x01(\x0b\x32\x19.videoimu.CameraInfo.Size\x12!\n\x19original_intrinsic_params\x18\r \x03(\x02\x1a%\n\x04Size\x12\r\n\x05width\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05"E\n\x10\x46ocusCalibration\x12\x10\n\x0cUNCALIBRATED\x10\x00\x12\x0f\n\x0b\x41PPROXIMATE\x10\x01\x12\x0e\n\nCALIBRATED\x10\x02",\n\x0fTimestampSource\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08REALTIME\x10\x01"E\n\x11LensPoseReference\x12\x12\n\x0ePRIMARY_CAMERA\x10\x00\x12\r\n\tGYROSCOPE\x10\x01\x12\r\n\tUNDEFINED\x10\x02";\n\x15VideoFrameToTimestamp\x12\x0f\n\x07time_us\x18\x01 \x01(\x03\x12\x11\n\tframe_nbr\x18\x02 \x01(\x03"\x82\x03\n\x12VideoFrameMetaData\x12\x0f\n\x07time_ns\x18\x01 \x01(\x03\x12\x14\n\x0c\x66rame_number\x18\x02 \x01(\x03\x12\x18\n\x10\x65xposure_time_ns\x18\x03 \x01(\x03\x12\x19\n\x11\x66rame_duration_ns\x18\x04 \x01(\x03\x12\x18\n\x10\x66rame_readout_ns\x18\x05 \x01(\x03\x12\x0b\n\x03iso\x18\x06 \x01(\x05\x12\x17\n\x0f\x66ocal_length_mm\x18\x07 \x01(\x02\x12\x1c\n\x14\x65st_focal_length_pix\x18\x08 \x01(\x02\x12\x1f\n\x17\x66ocus_distance_diopters\x18\t \x01(\x02\x12;\n\x0bOIS_samples\x18\n \x03(\x0b\x32&.videoimu.VideoFrameMetaData.OISSample\x12\x14\n\x0c\x66ocus_locked\x18\x0b \x01(\x08\x1a>\n\tOISSample\x12\x0f\n\x07time_ns\x18\x01 \x01(\x03\x12\x0f\n\x07x_shift\x18\x02 \x01(\x02\x12\x0f\n\x07y_shift\x18\x03 \x01(\x02"\x90\x01\n\x07IMUInfo\x12\x11\n\tgyro_info\x18\x01 \x01(\t\x12\x17\n\x0fgyro_resolution\x18\x02 \x01(\x02\x12\x12\n\naccel_info\x18\x03 \x01(\t\x12\x18\n\x10\x61\x63\x63\x65l_resolution\x18\x04 \x01(\x02\x12\x18\n\x10sample_frequency\x18\x05 \x01(\x02\x12\x11\n\tplacement\x18\x06 \x03(\x02"\x81\x02\n\x07IMUData\x12\x0f\n\x07time_ns\x18\x01 \x01(\x03\x12\x0c\n\x04gyro\x18\x02 \x03(\x02\x12\x12\n\ngyro_drift\x18\x03 \x03(\x02\x12\r\n\x05\x61\x63\x63\x65l\x18\x04 \x03(\x02\x12\x12\n\naccel_bias\x18\x05 \x03(\x02\x12\x31\n\rgyro_accuracy\x18\x06 \x01(\x0e\x32\x1a.videoimu.IMUData.Accuracy\x12\x32\n\x0e\x61\x63\x63\x65l_accuracy\x18\x07 \x01(\x0e\x32\x1a.videoimu.IMUData.Accuracy"9\n\x08\x41\x63\x63uracy\x12\x0e\n\nUNRELIABLE\x10\x00\x12\x07\n\x03LOW\x10\x01\x12\n\n\x06MEDIUM\x10\x02\x12\x08\n\x04HIGH\x10\x03"\xde\x01\n\x10VideoCaptureData\x12(\n\x04time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12)\n\x0b\x63\x61mera_meta\x18\x02 \x01(\x0b\x32\x14.videoimu.CameraInfo\x12#\n\x08imu_meta\x18\x03 \x01(\x0b\x32\x11.videoimu.IMUInfo\x12\x1e\n\x03imu\x18\x04 \x03(\x0b\x32\x11.videoimu.IMUData\x12\x30\n\nvideo_meta\x18\x05 \x03(\x0b\x32\x1c.videoimu.VideoFrameMetaData"\xfd\x01\n\x0eMessageWrapper\x12+\n\x0b\x63\x61mera_meta\x18\x01 \x01(\x0b\x32\x14.videoimu.CameraInfoH\x00\x12%\n\x08imu_data\x18\x02 \x01(\x0b\x32\x11.videoimu.IMUDataH\x00\x12%\n\x08imu_meta\x18\x03 \x01(\x0b\x32\x11.videoimu.IMUInfoH\x00\x12\x32\n\nframe_meta\x18\x04 \x01(\x0b\x32\x1c.videoimu.VideoFrameMetaDataH\x00\x12\x35\n\nframe_time\x18\x05 \x01(\x0b\x32\x1f.videoimu.VideoFrameToTimestampH\x00\x42\x05\n\x03msgB.\n\x1bse.lth.math.videoimucaptureB\x0fRecordingProtosb\x06proto3',
    dependencies=[
        google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,
    ],
)


_CAMERAINFO_FOCUSCALIBRATION = _descriptor.EnumDescriptor(
    name="FocusCalibration",
    full_name="videoimu.CameraInfo.FocusCalibration",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNCALIBRATED",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="APPROXIMATE",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="CALIBRATED",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=703,
    serialized_end=772,
)
_sym_db.RegisterEnumDescriptor(_CAMERAINFO_FOCUSCALIBRATION)

_CAMERAINFO_TIMESTAMPSOURCE = _descriptor.EnumDescriptor(
    name="TimestampSource",
    full_name="videoimu.CameraInfo.TimestampSource",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNKNOWN",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="REALTIME",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=774,
    serialized_end=818,
)
_sym_db.RegisterEnumDescriptor(_CAMERAINFO_TIMESTAMPSOURCE)

_CAMERAINFO_LENSPOSEREFERENCE = _descriptor.EnumDescriptor(
    name="LensPoseReference",
    full_name="videoimu.CameraInfo.LensPoseReference",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="PRIMARY_CAMERA",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="GYROSCOPE",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="UNDEFINED",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=820,
    serialized_end=889,
)
_sym_db.RegisterEnumDescriptor(_CAMERAINFO_LENSPOSEREFERENCE)

_IMUDATA_ACCURACY = _descriptor.EnumDescriptor(
    name="Accuracy",
    full_name="videoimu.IMUData.Accuracy",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNRELIABLE",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="LOW",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="MEDIUM",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="HIGH",
            index=3,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=1689,
    serialized_end=1746,
)
_sym_db.RegisterEnumDescriptor(_IMUDATA_ACCURACY)


_CAMERAINFO_SIZE = _descriptor.Descriptor(
    name="Size",
    full_name="videoimu.CameraInfo.Size",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="width",
            full_name="videoimu.CameraInfo.Size.width",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="height",
            full_name="videoimu.CameraInfo.Size.height",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=664,
    serialized_end=701,
)

_CAMERAINFO = _descriptor.Descriptor(
    name="CameraInfo",
    full_name="videoimu.CameraInfo",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="intrinsic_params",
            full_name="videoimu.CameraInfo.intrinsic_params",
            index=0,
            number=1,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="distortion_params",
            full_name="videoimu.CameraInfo.distortion_params",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="optical_image_stabilization",
            full_name="videoimu.CameraInfo.optical_image_stabilization",
            index=2,
            number=3,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="video_stabilization",
            full_name="videoimu.CameraInfo.video_stabilization",
            index=3,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="distortion_correction",
            full_name="videoimu.CameraInfo.distortion_correction",
            index=4,
            number=10,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="sensor_orientation",
            full_name="videoimu.CameraInfo.sensor_orientation",
            index=5,
            number=14,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="focus_calibration",
            full_name="videoimu.CameraInfo.focus_calibration",
            index=6,
            number=5,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp_source",
            full_name="videoimu.CameraInfo.timestamp_source",
            index=7,
            number=6,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="lens_pose_reference",
            full_name="videoimu.CameraInfo.lens_pose_reference",
            index=8,
            number=7,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="lens_pose_rotation",
            full_name="videoimu.CameraInfo.lens_pose_rotation",
            index=9,
            number=8,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="lens_pose_translation",
            full_name="videoimu.CameraInfo.lens_pose_translation",
            index=10,
            number=9,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="resolution",
            full_name="videoimu.CameraInfo.resolution",
            index=11,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="pre_correction_active_array_size",
            full_name="videoimu.CameraInfo.pre_correction_active_array_size",
            index=12,
            number=12,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="original_intrinsic_params",
            full_name="videoimu.CameraInfo.original_intrinsic_params",
            index=13,
            number=13,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _CAMERAINFO_SIZE,
    ],
    enum_types=[
        _CAMERAINFO_FOCUSCALIBRATION,
        _CAMERAINFO_TIMESTAMPSOURCE,
        _CAMERAINFO_LENSPOSEREFERENCE,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=63,
    serialized_end=889,
)


_VIDEOFRAMETOTIMESTAMP = _descriptor.Descriptor(
    name="VideoFrameToTimestamp",
    full_name="videoimu.VideoFrameToTimestamp",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="time_us",
            full_name="videoimu.VideoFrameToTimestamp.time_us",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="frame_nbr",
            full_name="videoimu.VideoFrameToTimestamp.frame_nbr",
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=891,
    serialized_end=950,
)


_VIDEOFRAMEMETADATA_OISSAMPLE = _descriptor.Descriptor(
    name="OISSample",
    full_name="videoimu.VideoFrameMetaData.OISSample",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="time_ns",
            full_name="videoimu.VideoFrameMetaData.OISSample.time_ns",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="x_shift",
            full_name="videoimu.VideoFrameMetaData.OISSample.x_shift",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="y_shift",
            full_name="videoimu.VideoFrameMetaData.OISSample.y_shift",
            index=2,
            number=3,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1277,
    serialized_end=1339,
)

_VIDEOFRAMEMETADATA = _descriptor.Descriptor(
    name="VideoFrameMetaData",
    full_name="videoimu.VideoFrameMetaData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="time_ns",
            full_name="videoimu.VideoFrameMetaData.time_ns",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="frame_number",
            full_name="videoimu.VideoFrameMetaData.frame_number",
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="exposure_time_ns",
            full_name="videoimu.VideoFrameMetaData.exposure_time_ns",
            index=2,
            number=3,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="frame_duration_ns",
            full_name="videoimu.VideoFrameMetaData.frame_duration_ns",
            index=3,
            number=4,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="frame_readout_ns",
            full_name="videoimu.VideoFrameMetaData.frame_readout_ns",
            index=4,
            number=5,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="iso",
            full_name="videoimu.VideoFrameMetaData.iso",
            index=5,
            number=6,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="focal_length_mm",
            full_name="videoimu.VideoFrameMetaData.focal_length_mm",
            index=6,
            number=7,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="est_focal_length_pix",
            full_name="videoimu.VideoFrameMetaData.est_focal_length_pix",
            index=7,
            number=8,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="focus_distance_diopters",
            full_name="videoimu.VideoFrameMetaData.focus_distance_diopters",
            index=8,
            number=9,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="OIS_samples",
            full_name="videoimu.VideoFrameMetaData.OIS_samples",
            index=9,
            number=10,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="focus_locked",
            full_name="videoimu.VideoFrameMetaData.focus_locked",
            index=10,
            number=11,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _VIDEOFRAMEMETADATA_OISSAMPLE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=953,
    serialized_end=1339,
)


_IMUINFO = _descriptor.Descriptor(
    name="IMUInfo",
    full_name="videoimu.IMUInfo",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="gyro_info",
            full_name="videoimu.IMUInfo.gyro_info",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gyro_resolution",
            full_name="videoimu.IMUInfo.gyro_resolution",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="accel_info",
            full_name="videoimu.IMUInfo.accel_info",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="accel_resolution",
            full_name="videoimu.IMUInfo.accel_resolution",
            index=3,
            number=4,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="sample_frequency",
            full_name="videoimu.IMUInfo.sample_frequency",
            index=4,
            number=5,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="placement",
            full_name="videoimu.IMUInfo.placement",
            index=5,
            number=6,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1342,
    serialized_end=1486,
)


_IMUDATA = _descriptor.Descriptor(
    name="IMUData",
    full_name="videoimu.IMUData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="time_ns",
            full_name="videoimu.IMUData.time_ns",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gyro",
            full_name="videoimu.IMUData.gyro",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gyro_drift",
            full_name="videoimu.IMUData.gyro_drift",
            index=2,
            number=3,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="accel",
            full_name="videoimu.IMUData.accel",
            index=3,
            number=4,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="accel_bias",
            full_name="videoimu.IMUData.accel_bias",
            index=4,
            number=5,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gyro_accuracy",
            full_name="videoimu.IMUData.gyro_accuracy",
            index=5,
            number=6,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="accel_accuracy",
            full_name="videoimu.IMUData.accel_accuracy",
            index=6,
            number=7,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[
        _IMUDATA_ACCURACY,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1489,
    serialized_end=1746,
)


_VIDEOCAPTUREDATA = _descriptor.Descriptor(
    name="VideoCaptureData",
    full_name="videoimu.VideoCaptureData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="time",
            full_name="videoimu.VideoCaptureData.time",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="camera_meta",
            full_name="videoimu.VideoCaptureData.camera_meta",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="imu_meta",
            full_name="videoimu.VideoCaptureData.imu_meta",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="imu",
            full_name="videoimu.VideoCaptureData.imu",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="video_meta",
            full_name="videoimu.VideoCaptureData.video_meta",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1749,
    serialized_end=1971,
)


_MESSAGEWRAPPER = _descriptor.Descriptor(
    name="MessageWrapper",
    full_name="videoimu.MessageWrapper",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="camera_meta",
            full_name="videoimu.MessageWrapper.camera_meta",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="imu_data",
            full_name="videoimu.MessageWrapper.imu_data",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="imu_meta",
            full_name="videoimu.MessageWrapper.imu_meta",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="frame_meta",
            full_name="videoimu.MessageWrapper.frame_meta",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="frame_time",
            full_name="videoimu.MessageWrapper.frame_time",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="msg",
            full_name="videoimu.MessageWrapper.msg",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1974,
    serialized_end=2227,
)

_CAMERAINFO_SIZE.containing_type = _CAMERAINFO
_CAMERAINFO.fields_by_name["focus_calibration"].enum_type = _CAMERAINFO_FOCUSCALIBRATION
_CAMERAINFO.fields_by_name["timestamp_source"].enum_type = _CAMERAINFO_TIMESTAMPSOURCE
_CAMERAINFO.fields_by_name[
    "lens_pose_reference"
].enum_type = _CAMERAINFO_LENSPOSEREFERENCE
_CAMERAINFO.fields_by_name["resolution"].message_type = _CAMERAINFO_SIZE
_CAMERAINFO.fields_by_name[
    "pre_correction_active_array_size"
].message_type = _CAMERAINFO_SIZE
_CAMERAINFO_FOCUSCALIBRATION.containing_type = _CAMERAINFO
_CAMERAINFO_TIMESTAMPSOURCE.containing_type = _CAMERAINFO
_CAMERAINFO_LENSPOSEREFERENCE.containing_type = _CAMERAINFO
_VIDEOFRAMEMETADATA_OISSAMPLE.containing_type = _VIDEOFRAMEMETADATA
_VIDEOFRAMEMETADATA.fields_by_name[
    "OIS_samples"
].message_type = _VIDEOFRAMEMETADATA_OISSAMPLE
_IMUDATA.fields_by_name["gyro_accuracy"].enum_type = _IMUDATA_ACCURACY
_IMUDATA.fields_by_name["accel_accuracy"].enum_type = _IMUDATA_ACCURACY
_IMUDATA_ACCURACY.containing_type = _IMUDATA
_VIDEOCAPTUREDATA.fields_by_name[
    "time"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_VIDEOCAPTUREDATA.fields_by_name["camera_meta"].message_type = _CAMERAINFO
_VIDEOCAPTUREDATA.fields_by_name["imu_meta"].message_type = _IMUINFO
_VIDEOCAPTUREDATA.fields_by_name["imu"].message_type = _IMUDATA
_VIDEOCAPTUREDATA.fields_by_name["video_meta"].message_type = _VIDEOFRAMEMETADATA
_MESSAGEWRAPPER.fields_by_name["camera_meta"].message_type = _CAMERAINFO
_MESSAGEWRAPPER.fields_by_name["imu_data"].message_type = _IMUDATA
_MESSAGEWRAPPER.fields_by_name["imu_meta"].message_type = _IMUINFO
_MESSAGEWRAPPER.fields_by_name["frame_meta"].message_type = _VIDEOFRAMEMETADATA
_MESSAGEWRAPPER.fields_by_name["frame_time"].message_type = _VIDEOFRAMETOTIMESTAMP
_MESSAGEWRAPPER.oneofs_by_name["msg"].fields.append(
    _MESSAGEWRAPPER.fields_by_name["camera_meta"]
)
_MESSAGEWRAPPER.fields_by_name[
    "camera_meta"
].containing_oneof = _MESSAGEWRAPPER.oneofs_by_name["msg"]
_MESSAGEWRAPPER.oneofs_by_name["msg"].fields.append(
    _MESSAGEWRAPPER.fields_by_name["imu_data"]
)
_MESSAGEWRAPPER.fields_by_name[
    "imu_data"
].containing_oneof = _MESSAGEWRAPPER.oneofs_by_name["msg"]
_MESSAGEWRAPPER.oneofs_by_name["msg"].fields.append(
    _MESSAGEWRAPPER.fields_by_name["imu_meta"]
)
_MESSAGEWRAPPER.fields_by_name[
    "imu_meta"
].containing_oneof = _MESSAGEWRAPPER.oneofs_by_name["msg"]
_MESSAGEWRAPPER.oneofs_by_name["msg"].fields.append(
    _MESSAGEWRAPPER.fields_by_name["frame_meta"]
)
_MESSAGEWRAPPER.fields_by_name[
    "frame_meta"
].containing_oneof = _MESSAGEWRAPPER.oneofs_by_name["msg"]
_MESSAGEWRAPPER.oneofs_by_name["msg"].fields.append(
    _MESSAGEWRAPPER.fields_by_name["frame_time"]
)
_MESSAGEWRAPPER.fields_by_name[
    "frame_time"
].containing_oneof = _MESSAGEWRAPPER.oneofs_by_name["msg"]
DESCRIPTOR.message_types_by_name["CameraInfo"] = _CAMERAINFO
DESCRIPTOR.message_types_by_name["VideoFrameToTimestamp"] = _VIDEOFRAMETOTIMESTAMP
DESCRIPTOR.message_types_by_name["VideoFrameMetaData"] = _VIDEOFRAMEMETADATA
DESCRIPTOR.message_types_by_name["IMUInfo"] = _IMUINFO
DESCRIPTOR.message_types_by_name["IMUData"] = _IMUDATA
DESCRIPTOR.message_types_by_name["VideoCaptureData"] = _VIDEOCAPTUREDATA
DESCRIPTOR.message_types_by_name["MessageWrapper"] = _MESSAGEWRAPPER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CameraInfo = _reflection.GeneratedProtocolMessageType(
    "CameraInfo",
    (_message.Message,),
    {
        "Size": _reflection.GeneratedProtocolMessageType(
            "Size",
            (_message.Message,),
            {
                "DESCRIPTOR": _CAMERAINFO_SIZE,
                "__module__": "recording_pb2"
                # @@protoc_insertion_point(class_scope:videoimu.CameraInfo.Size)
            },
        ),
        "DESCRIPTOR": _CAMERAINFO,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.CameraInfo)
    },
)
_sym_db.RegisterMessage(CameraInfo)
_sym_db.RegisterMessage(CameraInfo.Size)

VideoFrameToTimestamp = _reflection.GeneratedProtocolMessageType(
    "VideoFrameToTimestamp",
    (_message.Message,),
    {
        "DESCRIPTOR": _VIDEOFRAMETOTIMESTAMP,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.VideoFrameToTimestamp)
    },
)
_sym_db.RegisterMessage(VideoFrameToTimestamp)

VideoFrameMetaData = _reflection.GeneratedProtocolMessageType(
    "VideoFrameMetaData",
    (_message.Message,),
    {
        "OISSample": _reflection.GeneratedProtocolMessageType(
            "OISSample",
            (_message.Message,),
            {
                "DESCRIPTOR": _VIDEOFRAMEMETADATA_OISSAMPLE,
                "__module__": "recording_pb2"
                # @@protoc_insertion_point(class_scope:videoimu.VideoFrameMetaData.OISSample)
            },
        ),
        "DESCRIPTOR": _VIDEOFRAMEMETADATA,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.VideoFrameMetaData)
    },
)
_sym_db.RegisterMessage(VideoFrameMetaData)
_sym_db.RegisterMessage(VideoFrameMetaData.OISSample)

IMUInfo = _reflection.GeneratedProtocolMessageType(
    "IMUInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _IMUINFO,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.IMUInfo)
    },
)
_sym_db.RegisterMessage(IMUInfo)

IMUData = _reflection.GeneratedProtocolMessageType(
    "IMUData",
    (_message.Message,),
    {
        "DESCRIPTOR": _IMUDATA,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.IMUData)
    },
)
_sym_db.RegisterMessage(IMUData)

VideoCaptureData = _reflection.GeneratedProtocolMessageType(
    "VideoCaptureData",
    (_message.Message,),
    {
        "DESCRIPTOR": _VIDEOCAPTUREDATA,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.VideoCaptureData)
    },
)
_sym_db.RegisterMessage(VideoCaptureData)

MessageWrapper = _reflection.GeneratedProtocolMessageType(
    "MessageWrapper",
    (_message.Message,),
    {
        "DESCRIPTOR": _MESSAGEWRAPPER,
        "__module__": "recording_pb2"
        # @@protoc_insertion_point(class_scope:videoimu.MessageWrapper)
    },
)
_sym_db.RegisterMessage(MessageWrapper)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



"""The IMU video image provider provides images from a camera with an integrated IMU together with the IMU data."""
from typing import List, Generator, Tuple

import cv2
import pandas as pd

from sevivi.log import logger
from .video_provider import VideoImageProvider
from ..dimensions import Dimensions
from ...config import find_matching_columns

logger = logger.getChild("plain_video_image_provider")


class ImuCameraImageProvider(VideoImageProvider):
    """The IMU video image provider provides images from a camera with an integrated IMU together with the IMU data."""

    def __init__(self, video_path: str, imu_df_path: str):
        self.__video = cv2.VideoCapture(video_path)
        self.data: pd.DataFrame = pd.read_csv(
            imu_df_path, index_col=0, parse_dates=True
        )

    def get_sync_dataframe(self, column_names: List[str]) -> pd.DataFrame:
        """
        Returns the selected portion of the IMU data.
        """
        matching_columns = find_matching_columns(self.data, column_names)
        return self.data[matching_columns]

    def images(self) -> Generator[Tuple[pd.Timestamp, bytes], None, None]:
        """Generate the images to be shown together with their timestamps"""
        while self.__video.isOpened():
            frame_exists, frame = self.__video.read()
            if frame_exists:
                ts = self.__video.get(cv2.CAP_PROP_POS_MSEC)
                yield pd.to_datetime(ts, unit="ms"), frame
            else:
                break

    def get_image_count(self) -> int:
        """Get the number of images that will be rendered"""
        return int(self.__video.get(cv2.CAP_PROP_FRAME_COUNT))

    def get_dimensions(self) -> Dimensions:
        """Get the dimensions of the source video in pixels."""
        return Dimensions(
            w=int(self.__video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            h=int(self.__video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )



"""Contains helpers to instantiate Video- and GraphImageProviders for the CLI"""

from typing import Dict, List

import pandas as pd

from sevivi.config import (
    CameraImuVideoConfig,
    VideoImuCaptureAppVideoConfig,
    KinectVideoConfig,
    RawVideoConfig,
    OpenPoseVideoConfig,
    VideoConfig,
    Config,
    RenderConfig,
    SensorConfig,
)
from sevivi.image_provider import (
    AzureProvider,
    GraphImageProvider,
    ImuCameraImageProvider,
    VideoImageProvider,
    PlainVideoImageProvider,
    VideoImuCaptureAppImageProvider,
)
from .video_renderer import VideoRenderer


def video_renderer_from_csv_files(config: Config) -> VideoRenderer:
    """Instantiate a VideoRenderer for Configs where all data is stored as .csv or .csv.gz"""
    missing_files = config.get_missing_files()
    if len(missing_files) > 0:
        raise FileNotFoundError(
            f"Your configuration points to the following files that are not available: {missing_files}"
        )

    video_provider = instantiate_video_provider(config.video_config)
    graph_providers = instantiate_graph_providers(
        config.sensor_configs, config.render_config
    )
    return VideoRenderer(config.render_config, video_provider, graph_providers)


def instantiate_graph_providers(
    sensor_configs: Dict[str, SensorConfig], render_config: RenderConfig
) -> List[GraphImageProvider]:
    """
    Instantiate a GraphImageProvider for each config.

    You must instantiate your GraphImageProviders manually if the data isn't stored as CSV or CSV.GZ where the
    first column is a DatetimeIndex
    """
    result = []

    for sc in sensor_configs.values():
        data = pd.read_csv(sc.path, index_col=0, parse_dates=True)
        result.append(GraphImageProvider(data, sc, render_config.plotting_method))

    return result


def instantiate_video_provider(video_config: VideoConfig) -> VideoImageProvider:
    """Instantiate the appropriate VideoImageProvider subclass for a given VideoConfig"""
    if isinstance(video_config, VideoImuCaptureAppVideoConfig):
        return VideoImuCaptureAppImageProvider(video_config.path, video_config.imu_path)
    elif isinstance(video_config, CameraImuVideoConfig):
        return ImuCameraImageProvider(video_config.path, video_config.imu_path)
    elif isinstance(video_config, KinectVideoConfig):
        return AzureProvider(
            video_config.path,
            video_config.skeleton_path_3d,
            video_config.skeleton_path_2d,
        )
    elif isinstance(video_config, RawVideoConfig):
        return PlainVideoImageProvider(video_config.path)
    elif isinstance(video_config, OpenPoseVideoConfig):
        raise NotImplementedError("OpenPoseVideoConfig Not Implemented")
    else:
        raise RuntimeError(
            f"Unknown instance type of VideoConfig: {type(video_config)}"
        )



from .config_types.plotting_method import PlottingMethod
from .config_types.stacking_direction import StackingDirection
from .config_types.video_config import VideoConfig
from .config_types.config import Config, RenderConfig
from .column_matching import find_matching_columns, get_graph_groups

from .config_types.sensor_config import (
    SensorConfig,
    ManuallySynchronizedSensorConfig,
    JointSynchronizedSensorConfig,
    ImuSynchronizedSensorConfig,
)

from .config_types.video_config import (
    CameraImuVideoConfig,
    KinectVideoConfig,
    RawVideoConfig,
    OpenPoseVideoConfig,
    VideoImuCaptureAppVideoConfig,
)

from .config_reader import read_configs



import pandas as pd

from sevivi import video_renderer_from_csv_files, read_configs

import pytest


def test_get_sync_dataframe_wrong_type(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/kinect_sync_walking.toml",))
    ap = video_renderer_from_csv_files(config).video_provider

    with pytest.raises(TypeError):
        ap.get_sync_dataframe(5)


def test_get_sync_dataframe_wrong_joint(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/kinect_sync_walking.toml",))
    ap = video_renderer_from_csv_files(config).video_provider

    with pytest.raises(KeyError):
        ap.get_sync_dataframe("Unknown_Joint")


def test_get_sync_dataframe_wrong_joint_in_list(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/kinect_sync_walking.toml",))
    ap = video_renderer_from_csv_files(config).video_provider

    with pytest.raises(KeyError):
        ap.get_sync_dataframe(["PELVIS (x)", "UNKNOWN_JOINT"])


def test_get_sync_dataframe_result_from_list(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/kinect_sync_walking.toml",))
    ap = video_renderer_from_csv_files(config).video_provider
    sync_axes = ["PELVIS (x)", "PELVIS (y)", "PELVIS (z)"]

    df = ap.get_sync_dataframe(sync_axes)
    assert isinstance(df, pd.DataFrame)
    assert len(df.columns) == 3
    assert all(elem in df.columns for elem in sync_axes)


def test_get_sync_dataframe_result_from_str(run_in_repo_root):
    config = read_configs(("test_files/test-data-configs/kinect_sync_walking.toml",))
    ap = video_renderer_from_csv_files(config).video_provider

    sync_joint = "PELVIS (x)"
    df = ap.get_sync_dataframe(sync_joint)
    assert isinstance(df, pd.DataFrame)
    assert len(df.columns) == 1
    assert df.columns[0] == sync_joint



from typing import List, Optional, Dict, Union

import pandas as pd


def get_graph_groups(
    df: pd.DataFrame, patterns: Optional[List[str]]
) -> Dict[str, List[str]]:
    if patterns is None:
        return {col: [col] for col in df.columns}
    else:
        return {pattern: find_matching_columns(df, pattern) for pattern in patterns}


def find_matching_columns(
    df: pd.DataFrame, pattern: Union[str, List[str]]
) -> List[str]:
    if isinstance(pattern, str):
        return [col for col in df.columns if pattern in col]
    else:
        column_set = set(df.columns)
        desired_column_set = set(pattern)
        if len(desired_column_set - column_set) > 0:
            raise ValueError(f"Unknown columns {desired_column_set - column_set}")
        return pattern



import os
from dataclasses import dataclass
from typing import List


@dataclass
class VideoConfig:
    """
    Base class for video configs.
    """

    path: str = None
    """Path to the video file. May be absolute or relative"""

    def get_missing_files(self) -> List[str]:
        """Returns a list of all missing files for this config"""
        if not os.path.isfile(self.path):
            return [self.path]
        return []


@dataclass
class RawVideoConfig(VideoConfig):
    """Can be used to fully configure raw video files."""


@dataclass
class KinectVideoConfig(VideoConfig):
    """Specify Kinect input video"""

    skeleton_path_3d: str = None
    """Path to the 3D skeleton data"""
    skeleton_path_2d: str = None
    """Path to the 2D skeleton data"""

    def get_missing_files(self) -> List[str]:
        """Returns a list of all missing files for this config"""
        missing_files = super().get_missing_files()
        if not os.path.isfile(self.skeleton_path_3d):
            missing_files.append(self.skeleton_path_3d)
        if not os.path.isfile(self.skeleton_path_2d):
            missing_files.append(self.skeleton_path_2d)
        return missing_files


@dataclass
class CameraImuVideoConfig(VideoConfig):
    """Specify camera IMU input video"""

    imu_path: str = None
    """Path to the IMU data corresponding to the camera"""

    def get_missing_files(self) -> List[str]:
        """Returns true if all files for this config exist"""
        missing_files = super().get_missing_files()
        if not os.path.isfile(self.imu_path):
            missing_files.append(self.imu_path)
        return missing_files


@dataclass
class VideoImuCaptureAppVideoConfig(CameraImuVideoConfig):
    """Specify input from the Video IMU Capture App"""

    pass


@dataclass
class OpenPoseVideoConfig(VideoConfig):
    """Specify OpenPose input video"""



from enum import Enum


class PlottingMethod(Enum):
    """The plotting method to use"""

    MOVING_VERTICAL_LINE = 0
    """
    The MOVING_VERTICAL_LINE plotting method is the fastest.
    The entire graph is shown for the entire video.
    A vertical line is moved through the graph to show the current moment in time.
    """
    PUSH_IN = 1
    """
    The PUSH_IN plotting method adds new sensor data on the right edge of the graph.
    This compresses the data on the left. The most recent data is always on the rightmost edge.
    """



joint_definition = [
    ["HEAD", "NECK"],
    ["NECK", "SPINE_CHEST"],
    ["SPINE_CHEST", "SPINE_NAVEL"],
    ["SPINE_NAVEL", "PELVIS"],
    ["PELVIS", "HIP_RIGHT"],
    ["HIP_RIGHT", "KNEE_RIGHT"],
    ["KNEE_RIGHT", "ANKLE_RIGHT"],
    ["ANKLE_RIGHT", "FOOT_RIGHT"],
    ["PELVIS", "HIP_LEFT"],
    ["HIP_LEFT", "KNEE_LEFT"],
    ["KNEE_LEFT", "ANKLE_LEFT"],
    ["ANKLE_LEFT", "FOOT_LEFT"],
    ["SPINE_CHEST", "CLAVICLE_LEFT"],
    ["CLAVICLE_LEFT", "SHOULDER_LEFT"],
    ["SHOULDER_LEFT", "ELBOW_LEFT"],
    ["ELBOW_LEFT", "WRIST_LEFT"],
    ["WRIST_LEFT", "HAND_LEFT"],
    ["HAND_LEFT", "HANDTIP_LEFT"],
    ["WRIST_LEFT", "THUMB_LEFT"],
    ["SPINE_CHEST", "CLAVICLE_RIGHT"],
    ["CLAVICLE_RIGHT", "SHOULDER_RIGHT"],
    ["SHOULDER_RIGHT", "ELBOW_RIGHT"],
    ["ELBOW_RIGHT", "WRIST_RIGHT"],
    ["WRIST_RIGHT", "HAND_RIGHT"],
    ["HAND_RIGHT", "HANDTIP_RIGHT"],
    ["WRIST_RIGHT", "THUMB_RIGHT"],
]



import pytest

from sevivi.config import config_reader, RenderConfig
from sevivi.config.config_types.sensor_config import (
    ManuallySynchronizedSensorConfig,
    ImuSynchronizedSensorConfig,
    JointSynchronizedSensorConfig,
)
from sevivi.config.config_types.video_config import (
    CameraImuVideoConfig,
    KinectVideoConfig,
    RawVideoConfig,
    OpenPoseVideoConfig,
    VideoImuCaptureAppVideoConfig,
)
from sevivi.image_provider import (
    AzureProvider,
    PlainVideoImageProvider,
    ImuCameraImageProvider,
    VideoImuCaptureAppImageProvider,
)
from sevivi.video_renderer import VideoRenderer
from sevivi.video_renderer.instantiation_helpers import (
    instantiate_video_provider,
    instantiate_graph_providers,
    video_renderer_from_csv_files,
)


def test_instantiate_cam_imu_video_provider(run_in_repo_root):
    assert isinstance(
        instantiate_video_provider(
            CameraImuVideoConfig(
                path="test_files/videos/joint_synchronization_squatting.mp4",
                imu_path="test_files/skeletons/joint_synchronization_squatting/positions_3d.csv.gz",
            )
        ),
        ImuCameraImageProvider,
    )


def test_instantiate_smartphone_imu_video_provider(run_in_repo_root):
    assert isinstance(
        instantiate_video_provider(
            VideoImuCaptureAppVideoConfig(
                path="test_files/videos/imu_sync.mp4",
                imu_path="test_files/sensors/video_imu_capture_app/video_meta.pb3",
            )
        ),
        VideoImuCaptureAppImageProvider,
    )


def test_instantiate_kinect_video_provider(run_in_repo_root):
    assert isinstance(
        instantiate_video_provider(
            KinectVideoConfig(
                path="test_files/videos/joint_synchronization_squatting.mp4",
                skeleton_path_3d="test_files/skeletons/joint_synchronization_squatting/positions_3d.csv.gz",
            )
        ),
        AzureProvider,
    )


def test_instantiate_raw_video_provider():
    assert isinstance(
        instantiate_video_provider(RawVideoConfig()), PlainVideoImageProvider
    )


def test_instantiate_openpose_video_provider():
    with pytest.raises(NotImplementedError):
        instantiate_video_provider(OpenPoseVideoConfig())


def test_instantiate_unknown_video_provider():
    with pytest.raises(RuntimeError):
        # noinspection PyTypeChecker
        instantiate_video_provider(3)


def test_instantiate_graph_providers(run_in_repo_root):
    # noinspection PyTypeChecker
    configs = {
        "1": ManuallySynchronizedSensorConfig(
            path="test_files/sensors/imu_synchronization/LF.csv.gz"
        ),
        "2": ImuSynchronizedSensorConfig(
            path="test_files/sensors/imu_synchronization/LF.csv.gz"
        ),
        "3": JointSynchronizedSensorConfig(
            path="test_files/sensors/joint_synchronization_squatting/LF.csv.gz"
        ),
    }
    result = instantiate_graph_providers(configs, RenderConfig())
    assert len(result) == 3


def test_video_renderer_from_csv_files(run_in_repo_root):
    config = config_reader.read_configs(
        (
            "test_files/configs/basic_config.toml",
            "test_files/test-data-configs/kinect_sync_3d_hand_movement.toml",
        )
    )
    result = video_renderer_from_csv_files(config)
    assert isinstance(result, VideoRenderer)



from typing import List, Optional, Generator, Tuple
from collections import OrderedDict

import pandas as pd
import numpy as np
import matplotlib.colors
import os
import cv2

from .video_provider import VideoImageProvider
from ..dimensions import Dimensions
from .azure_connections import joint_definition as azure_joints


def get_interpolated_hsv_color_as_rgb(cur_value: float, max_value: float) -> np.ndarray:
    """Get a RGB color value based on the HSV color spectrum"""
    return matplotlib.colors.hsv_to_rgb([cur_value / max_value, 1, 1]) * 255


class AzureProvider(VideoImageProvider):
    """The Azure Provider provides images from a video or path with images along"""

    def __init__(
        self,
        video_path: str,
        joint_3d_df: str,
        joint_2d_df: Optional[str],
    ):
        if not os.path.exists(video_path):
            raise Exception(f"File {video_path} not found!")

        self.__video = cv2.VideoCapture(video_path)
        self.__joint_df_3d = self._drop_duplicate_body_indices_and_confidence_values(
            pd.read_csv(joint_3d_df, sep=";", index_col=0)
        )
        self.__joint_df_3d.index = pd.to_datetime(self.__joint_df_3d.index, unit="us")

        if joint_2d_df is not None:
            self.__joint_df_2d = (
                self._drop_duplicate_body_indices_and_confidence_values(
                    pd.read_csv(joint_2d_df, sep=";", index_col=0)
                )
            )
            self.__joint_df_2d.index = pd.to_datetime(
                self.__joint_df_2d.index, unit="us"
            )
            self.__skeleton_definition = self._get_skeleton_definition_as_tuple(
                self.__joint_df_2d
            )
        else:
            self.__joint_df_2d = None
            self.__skeleton_definition = None

    @staticmethod
    def _drop_duplicate_body_indices_and_confidence_values(
        df: pd.DataFrame,
    ) -> pd.DataFrame:
        """Drop body indices and confidence values from data frame"""
        df = df[[c for c in df.columns if "(c)" not in c]]
        body_idx_c = df["body_idx"].value_counts()
        df = df[df["body_idx"] == body_idx_c.index[body_idx_c.argmax()]]
        df = df.drop("body_idx", axis=1)
        return df

    @staticmethod
    def _get_skeleton_definition_as_tuple(df: pd.DataFrame) -> List[Tuple[int, int]]:
        """Returns the joint connection structure for the Azure Kinect skeleton"""
        joints = AzureProvider.get_joints_as_list(df)
        skeleton = []
        for j1, j2 in azure_joints:
            if j1 not in joints or j2 not in joints:
                continue
            skeleton.append((joints.index(j1), joints.index(j2)))

        return skeleton

    @staticmethod
    def get_joints_as_list(df: pd.DataFrame) -> List[str]:
        """Extracts all skeleton joints from (x,y,z) dataframe columns"""
        columns = [
            column.replace(column[column.find(" (") : column.find(")") + 1], "")
            for column in df.columns
        ]
        return list(OrderedDict.fromkeys(columns))

    def images(self) -> Generator[Tuple[pd.Timestamp, bytes], None, None]:
        """Delivers individual frames from Azure Kinect camera"""
        image_count = 0
        while self.__video.isOpened():
            frame_exists, frame = self.__video.read()
            if frame_exists:
                ts = self.__video.get(cv2.CAP_PROP_POS_MSEC)

                if self.__skeleton_definition is not None and image_count < len(
                    self.__joint_df_2d
                ):
                    skeleton_data = (
                        self.__joint_df_2d.iloc[image_count, :]
                        .to_numpy()
                        .reshape(-1, 2)
                    )
                    for joint_count, (j1, j2) in enumerate(self.__skeleton_definition):
                        p1 = tuple(int(x) for x in skeleton_data[j1])
                        p2 = tuple(int(x) for x in skeleton_data[j2])
                        color = get_interpolated_hsv_color_as_rgb(
                            joint_count, self.__joint_df_2d.shape[1]
                        )
                        frame = cv2.line(frame, p1, p2, color=color, thickness=9)

                yield pd.to_datetime(ts, unit="ms"), frame
                image_count += 1
            else:
                break

    def get_sync_dataframe(self, column_names: List[str]) -> Optional[pd.DataFrame]:
        """Returns a dataframe used for synchronization based on given joint names"""
        if type(column_names) is str:
            if column_names not in self.__joint_df_3d.columns:
                raise KeyError(
                    f"Sync joint name {column_names} not in Kinect dataframe."
                )
            else:
                return self.__joint_df_3d[column_names].to_frame()

        if type(column_names) is list:
            if not all(elem in self.__joint_df_3d.columns for elem in column_names):
                raise KeyError(f"Missing sync joints in Dataframe: {column_names}.")
            return self.__joint_df_3d[
                [col for col in column_names if col in self.__joint_df_3d.columns]
            ]

        raise TypeError(
            f"Wrong type given: {type(column_names)}, expected: List[str] or str"
        )

    def get_image_count(self) -> int:
        """Get the number of images that will be rendered"""
        return int(self.__video.get(cv2.CAP_PROP_FRAME_COUNT))

    def get_dimensions(self) -> Dimensions:
        """Get the dimensions of the source video in pixels"""
        return Dimensions(
            w=int(self.__video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            h=int(self.__video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sevivi',
 'sevivi.config',
 'sevivi.config.config_types',
 'sevivi.image_provider',
 'sevivi.image_provider.graph_provider',
 'sevivi.image_provider.video_provider',
 'sevivi.image_provider.video_provider.video_imu_capture_app',
 'sevivi.synchronizer',
 'sevivi.video_renderer']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.2,<9.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.3,<2.0.0',
 'opencv-python>=4.5.4.58,<5.0.0.0',
 'pandas>=1.3.4,<2.0.0',
 'read-protobuf>=0.1.1,<0.2.0',
 'scipy>=1.7.1,<2.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['sevivi = sevivi.main:run']}

setup_kwargs = {
    'name': 'sevivi',
    'version': '1.0.2',
    'description': 'Create a video with graphs synchronous to a source video',
    'long_description': None,
    'author': 'Justin Albert',
    'author_email': 'justin.albert@hpi.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hpi-dhc/sevivi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from .signal_processing import (
    resample_data,
    normalize_signal,
    calculate_magnitude,
    calculate_offset_in_seconds_using_cross_correlation,
    calculate_sampling_frequency_from_timestamps,
)


def get_synchronization_offset(
    video_sync_df: pd.DataFrame,
    sensor_sync_df: pd.DataFrame,
    use_gradient: bool,
    show_plots: bool = False,
) -> pd.Timedelta:
    """
    Get the temporal offset between the two given sensor dataframes.

    :param video_sync_df: the synchronization information from the video
    :param sensor_sync_df: the synchronization information from the sensor
    :param use_gradient: if true, the second derivation of the video synchronization data will be used. if false,
                         the raw data will be used.
    :param show_plots:  can enable debugging plots
    :return: a pd.Timedelta object that specifies how much the sensor_sync_df needs to be moved in time to align it with
            the video_sync_df
    """
    video_sf = calculate_sampling_frequency_from_timestamps(video_sync_df.index)
    sensor_sf = calculate_sampling_frequency_from_timestamps(sensor_sync_df.index)

    if use_gradient:
        video_acceleration = np.gradient(
            np.gradient(video_sync_df.to_numpy(), axis=0), axis=0
        )
    else:
        video_acceleration = video_sync_df.to_numpy()

    video_acceleration = resample_data(
        video_acceleration,
        current_sampling_rate=video_sf,
        new_sampling_rate=sensor_sf,
    )
    video_acceleration = normalize_signal(video_acceleration)
    video_acceleration = calculate_magnitude(video_acceleration)

    sensor_acceleration = normalize_signal(sensor_sync_df.to_numpy())
    sensor_acceleration = calculate_magnitude(sensor_acceleration)

    if show_plots:
        plt.close()
        plt.figure(1)
        plt.plot(video_acceleration, label="Kinect")
        plt.plot(sensor_acceleration, label="IMU")
        plt.xlabel("Time (s)")
        plt.ylabel("Acceleration Magnitude (normalized)")
        plt.legend()
        plt.show()

    shift = calculate_offset_in_seconds_using_cross_correlation(
        ref_signal=video_acceleration,
        target_signal=sensor_acceleration,
        sampling_frequency=sensor_sf,
    )

    if show_plots:
        plt.close()
        plt.figure(1)
        plt.plot(video_acceleration, label="Kinect")
        plt.plot(
            np.arange(len(sensor_acceleration)) + (sensor_sf * shift),
            sensor_acceleration,
            label="IMU",
        )
        plt.xlabel("Time (s)")
        plt.ylabel("Acceleration (normalized)")
        plt.legend()
        plt.show()

    return pd.Timedelta(seconds=shift)
